import { Component } from "./Component.js";

/**
 * Extracts an error message from a given error event.
 *
 * @param event - An event.
 */

function getErrorMessage(event: ErrorEvent | PromiseRejectionEvent): string {

	if(event instanceof ErrorEvent) {

		return event.message;

	}

	return (typeof event.reason === "string") ? event.reason : (event.reason as ErrorEvent).message;

}

/**
 * A viewport manager.
 */

export class ViewportManager implements Component, EventListenerObject {

	/**
	 * The viewport.
	 */

	private viewport: HTMLElement | null;

	/**
	 * A fullscreen button.
	 */

	private fullscreenButton: HTMLElement | null;

	constructor() {

		this.viewport = null;
		this.fullscreenButton = null;

	}

	/**
	 * Shows an error message.
	 *
	 * @param message - The error message.
	 */

	private showErrorMessage(message: string): void {

		if(this.viewport === null) {

			return;

		}

		const viewport = this.viewport;
		const canvas = viewport.querySelector("canvas");
		const warning = viewport.querySelector(".warning");
		const error = viewport.querySelector(".error");
		const p = error?.querySelector("p") as HTMLElement;

		if(p !== null) {

			p.innerText = message;

		}

		viewport?.classList.remove("loading");
		warning?.classList.add("hidden");
		canvas?.classList.add("hidden");
		error?.classList.remove("hidden");

	}

	/**
	 * Shows an epilepsy warning.
	 *
	 * @private
	 */

	showEpilepsyWarning() {

		if(this.viewport === null) {

			return;

		}

		const viewport = this.viewport;
		const canvas = viewport.querySelector("canvas");
		const warning = viewport.querySelector(".warning");
		const tp = viewport.querySelector(".tp");
		const a = warning?.querySelector("a");

		tp?.classList.toggle("hidden");
		canvas?.classList.toggle("hidden");
		warning?.classList.toggle("hidden");

		a?.addEventListener("click", (event) => {

			event.preventDefault();
			sessionStorage.setItem("epilepsy-warning", "1");
			warning?.classList.toggle("hidden");
			canvas?.classList.toggle("hidden");
			tp?.classList.toggle("hidden");

		});

	}

	/**
	 * Handles scroll events.
	 *
	 * @param event - The event.
	 */

	private handleScroll(event: Event): void {

		if(this.viewport !== null && this.viewport.matches(":hover")) {

			event.preventDefault();

		}

	}

	/**
	 * Toggles fullscreen mode.
	 */

	private toggleFullscreen(): void {

		if(document.fullscreenEnabled) {

			if(document.fullscreenElement !== null) {

				document.exitFullscreen().catch((e) => console.error(e));

			} else {

				this.viewport?.requestFullscreen().catch((e) => console.error(e));

			}

		}

	}

	handleEvent(event: Event): void {

		switch(event.type) {

			case "wheel":
				this.handleScroll(event);
				break;

			case "click":
				this.toggleFullscreen();
				break;

			case "fullscreenchange":
				this.fullscreenButton?.classList.toggle("active");
				break;

			case "unhandledrejection":
				this.showErrorMessage(getErrorMessage(event as PromiseRejectionEvent));
				break;

		}

	}

	initialize(): void {

		const viewport = this.viewport = document.querySelector(".viewport");

		if(viewport === null) {

			return;

		}

		// Error Handling

		window.addEventListener("unhandledrejection", this);

		// Scrolling

		const main = document.getElementById("main");
		main?.addEventListener("wheel", this, { passive: false });

		// Fullscreen

		const fullscreenButton = this.fullscreenButton = viewport.querySelector(".fullscreen");
		fullscreenButton?.addEventListener("click", this);
		document.addEventListener("fullscreenchange", this);

		// Loading Animation

		const observer = new MutationObserver((mutationsList, observer) => {

			for(const mutation of mutationsList) {

				if(mutation.type === "childList") {

					// Stop the loading animation when something is added to the viewport.
					viewport.classList.remove("loading");

					// Show an epilepsy warning if applicable.
					const alreadyShown = (sessionStorage.getItem("epilepsy-warning") !== null);

					if(viewport.dataset.epilepsyWarning !== undefined && !alreadyShown) {

						this.showEpilepsyWarning();

					}

					observer.disconnect();

				}

			}

		});

		observer.observe(viewport, { childList: true });

	}

}
